#!/usr/bin/env python
import os
import sys
import time
import Tkinter

import Image
import ImageTk
import PIL.ImageOps as ImageOps

from ianno import common
from ianno import config
from ianno import dbannotations
from ianno import tagcloud


class Annotator(Tkinter.Frame):
	"""is a UI to edit annotations for images.

	It is constructed with an annotations instance and a list of files to 
	display.
	"""

	trashDir = os.path.join(os.environ["HOME"], ".ianno-trash")

	def __init__(self, master, annotations, flist):
		self.flist, self.annotations = flist, annotations
		self.current = None
		Tkinter.Frame.__init__(self, master)
		self.makeWidgets()
		self.doBindings()
		if self.flist:
			self.scrollist.select_set(0)
		def startUpdate(ev):
			self.unbind("<Map>")
			self.update_idletasks()
			self.wait_visibility()
			self.updateDisp()
		self.bind("<Map>", startUpdate)

	def _getSelection(self):
		newsel = self.scrollist.curselection()
		if newsel==():
			newsel = None
		else:
			newsel = int(newsel[0])
		return newsel

	def updateDisp(self):
		newsel = self._getSelection()
		if self.current!=newsel:
			self._changeImage(newsel)
		self.after(150, self.updateDisp)

	def next(self):
		if len(self.flist)-1>self.current and self.current is not None:
			newSel = self._getSelection()+1
			self.scrollist.select_clear(0, Tkinter.END)
			self._changeImage(newSel)
		return "break"

	def prev(self):
		if self.current is not None and self.current>0:
			newSel = self._getSelection()-1
			self.scrollist.select_clear(0, Tkinter.END)
			self._changeImage(newSel)
		return "break"

	def saveToDisk(self):
		self.annotations.save()

	def copyFromLastAnn(self):
		if self.lastAnn:
			self.curAnn.description = self.lastAnn.description
			self.curAnn.tags = self.lastAnn.tags.copy()
			self._displayAnn()

	def goToNextUntagged(self):
		cur = self.current+1
		try:
			while True:
				if self.annotations.get(self.flist[cur]) is None:
					self.scrollist.select_clear(0, Tkinter.END)
					self._changeImage(cur)
					break
				cur += 1
		except IndexError: # No untagged left
			pass

	def spawnFeh(self):
		if self.current is not None:
			os.system("feh -FZ '%s'&"%self.flist[self.current])

	def spawnGimp(self):
		if self.current is not None:
			os.system("gimp-remote '%s'&"%self.flist[self.current])

	def _ensureTrash(self):
		if not os.path.isdir(self.trashDir):
			os.mkdir(self.trashDir)

	def deleteCurrent(self):
		if self.current is not None:
			self._ensureTrash()
			key = self.flist[self.current]
			os.rename(key, os.path.join(self.trashDir, os.path.basename(key)))
			self.annotations.delete(key)
			del self.flist[self.current]
			self._fillScroller()
			nextIndex = self.current
			self.current = None
			if len(self.flist)>nextIndex:
				self._changeImage(nextIndex)
			elif self.flist:
				self._changeImage(len(self.flist)-1)
			else:
				self._changeImage(None)

	def _changeImage(self, newSelection):
		if self.current is not None:
			self._saveAnnotation()
		self.current = newSelection
		if self.current is not None:
			self.scrollist.select_set(self.current)
			self.scrollist.see(self.current)
		self._update()

	def _saveAnnotation(self):
		if self.current is None:
			return
		curDesc = self.textthing.get('1.0', Tkinter.END).strip()
		curTags = self.tagCloud.getSelectedItems()
		if curDesc or curTags:
			self.curAnn.description = unicode(curDesc.strip())
			self.curAnn.tags = curTags
			self.curAnn.updateDate()
			self.annotations.add(self.curAnn)
		else:
			self.annotations.delete(self.curAnn.key)

	def _update(self):
		self.tagCloud.clear()
		self.textthing.delete('1.0', Tkinter.END)
		if self.current is None:
			self.piczone.itemconfig(self.picId, image=None)
			self.textthing["state"] = Tkinter.DISABLED
		else:
			self._loadPic()
			self._updateAnn()

	def _updateAnn(self):
		self.lastAnn = getattr(self, "curAnn", None)
		self.curAnn = self.annotations.get(self.flist[self.current])
		if not self.curAnn:
			self.curAnn = common.Annotation(key=self.flist[self.current])
		self._displayAnn()
	
	def _displayAnn(self):
		self.textthing["state"] = Tkinter.NORMAL
		self.textthing.delete('1.0', Tkinter.END)
		self.textthing.insert(Tkinter.END, self.curAnn.description)
		self.tagCloud.select(self.curAnn.tags)
		
	def _loadPic(self):
		self.curim = ImageTk.PhotoImage(
			common.fitImage(Image.open(self.flist[self.current]),
				self.piczone))
		self.piczone.itemconfig(self.picId, image=self.curim)

	def quit(self):
		self._saveAnnotation()
		Tkinter.Frame.quit(self)
		return "break"

	def doBindings(self):
		self.bind_all("<Control-q>", lambda ev: self.quit())
		self.bind_all("<Control-space>", lambda ev: self.copyFromLastAnn())
		self.bind_all("<Control-Down>", lambda ev: self.next())
		self.bind_all("<Control-Up>", lambda ev: self.prev())
		self.bind_all("<Control-n>", lambda ev: self.goToNextUntagged())
		self.bind_all("<Control-s>", lambda ev: self.saveToDisk())
		self.textthing.bind("<Control-Return>", lambda ev: False)
		self.textthing.bind("<F1>", lambda ev: self.spawnFeh())
		self.textthing.bind("<F7>", lambda ev: self.deleteCurrent())
		self.textthing.bind("<F12>", lambda ev: self.spawnGimp())
		self.textthing.focus()

	def _fillScroller(self):
		self.scrollist.delete(0, Tkinter.END)
		for key in self.flist:
			self.scrollist.insert(Tkinter.END, key[-20:])

	def makeScroller(self):
		scrollframe = Tkinter.Frame(self)
		scrollbar = Tkinter.Scrollbar(scrollframe, orient=Tkinter.VERTICAL)
		self.scrollist = Tkinter.Listbox(scrollframe, width=20, height=30,
			selectmode=Tkinter.SINGLE, yscrollcommand=scrollbar.set,
			exportselection=0, font=("Helvetica",10))
		scrollbar.config(command=self.scrollist.yview)
		scrollbar.pack(side=Tkinter.RIGHT, fill=Tkinter.Y)
		self.scrollist.pack(side=Tkinter.LEFT, fill=Tkinter.BOTH, expand=1)
		self._fillScroller()
		return scrollframe

	def makeButtons(self):
		butframe = Tkinter.Frame(self)
		self.textthing = Tkinter.Text(butframe, height=3, font=("Helvetica",10))
		quitbut = Tkinter.Button(butframe, text="Quit", command=self.quit)
		quitbut.pack(side=Tkinter.LEFT)
		self.textthing.pack(side=Tkinter.LEFT)
		self.textthing["state"] = Tkinter.DISABLED
		return butframe

	def makeWidgets(self):
		self.tagCloud = tagcloud.TagCloudWithAdder(
			self, self.annotations.getTagFreqs())
		scrollframe = self.makeScroller()
		butframe = self.makeButtons()
		self.piczone = Tkinter.Canvas(self)
		self.piczone.config(width=640)
		self.picId = self.piczone.create_image(0, 0, anchor="nw")
		self.piczone.config(xscrollincrement=1, yscrollincrement=1)
		self.grid_propagate(0)

		scrollframe.grid(row=1, column=1, sticky=Tkinter.N+Tkinter.S)
		self.piczone.grid(row=1, column=2, sticky=Tkinter.NE+Tkinter.SW)
		self.tagCloud.grid(row=1, column=3, sticky=Tkinter.SE+Tkinter.NW)
		butframe.grid(row=2, column=1, columnspan=2)
		self.rowconfigure(1, weight=1)
		self.columnconfigure(3, weight=1)

	def getVolatiles(self):
		if False:
			yield


def updateDates(annotations):
	for ann in annotations:
		ann.updateDate()
		annotations.add(ann)
		sys.stdout.write(".")
		sys.stdout.flush()


def getFList():
	flist = [os.path.abspath(p) for p in sys.argv[1:]]
	flist.sort()
	return flist


def main():
	tkroot = Tkinter.Tk()
	tkroot.geometry(config.get("ianno", "windowGeometry"))
	tkroot.pack_propagate(0)
	anns = dbannotations.Annotations(os.path.join(os.environ["HOME"], 
		".ianno.sqlite"))
	u = Annotator(tkroot, anns, getFList())
	u.pack(expand=1,fill=Tkinter.BOTH)
	u.mainloop()
	anns.save()
	config.set("ianno", "windowGeometry", tkroot.geometry())
	for key, value in u.getVolatiles():
		config.set("ianno", key, value)
	config.save()



if __name__=="__main__":
	anns = dbannotations.Annotations(os.path.join(os.environ["HOME"], 
		".ianno.sqlite"))
	updateDates(anns)
	anns.save()
