"""
Basic definitions for the image tagger.
"""

import datetime
import os
import time

import Image

from ianno import EXIF

class Error(Exception):
	pass


class FormatError(Error):
	"""means a line line that cannot reasonably be deserialized was
	fed to Annotation
	"""


def fitImage(image, container):
	"""returns a PIL image fitting into container.

	container is some sort of Tkinter window.  If its winfo_width is not
	valid the function is called, image is returned unscaled.
	"""		
	iw, ih = image.size
	cw, ch = container.winfo_width(), container.winfo_height()
	if cw<2:
		return image
	ratio = max(iw/float(cw), ih/float(ch))
	newWidth, newHeight = int(iw/ratio), int(ih/ratio)
	return image.resize((newWidth, newHeight), Image.BILINEAR)


_dateTimeTag = "EXIF DateTimeOriginal"
def getFileDate(fPath):
	"""returns a "characteristic date" for fPath.

	If there's an exif info in fPath, that would be its creation date,
	otherwise it's what the file system says on the creation date.
	"""
	try:
		f = open(fPath)
		try:
			tagDict = EXIF.process_file(f, stop_tag=_dateTimeTag)
		except KeyError:  # some kind of internal EXIF error
			tagDict = {}
		f.close()
		if _dateTimeTag in tagDict:
			return datetime.datetime(*time.strptime(tagDict[_dateTimeTag].values,
				"%Y:%m:%d %H:%M:%S")[:6])
		else:
			return datetime.datetime.fromtimestamp(os.path.getctime(fPath))
	except IOError:  # file damaged or deleted
		return None


def _escape(s):
	return s.replace("\t", "\\t").replace("\n", "\\n")

def _unescape(s):
	return s.replace("\\t", "\t").replace("\\n", "\n")


class Annotation(object):
	"""is a container for information known about a file that serializes
	to text lines.
	"""
	def __init__(self, key=None, description="", tags=None, charDate=None):
		self.key, self.description = key, description
		self.tags = tags or set()
		self.charDate = charDate
		if not self.key:
			raise FormatError("Annotation without key")

	@classmethod
	def fromString(cls, serializedForm):
		parts = serializedForm.decode("utf-8").split("\t")
		charDate = None
		if len(parts)==3:
			pass
		elif len(parts)==4:
			if parts[3]!="None":
				charDate = datetime.datetime(*time.strptime(parts[3], 
					"%Y-%M-%DT%H:%m:%s"))
		else:
			raise FormatError("Badly serialized line: %s"%repr(serializedForm))
		key, description = map(_unescape, parts[:2])
		return cls(key=key, description=description,
			tags=set([s.strip() for s in parts[2].split(",") if s.strip()]),
			charDate = charDate)
	
	def toString(self):
		if self.charDate:
			charDate = self.charDate.isoformat()
		else:
			charDate = None
		return ("%s\t%s\t%s\t%s"%(
			_escape(self.key), _escape(self.description),
			",".join(self.tags), charDate)).encode("utf-8")
	
	def updateDate(self):
		self.charDate = getFileDate(self.key)

