"""
SQLite storage of annotations.
"""

from pysqlite2 import dbapi2 as sqlite

from ianno import common

_mainDDL = """CREATE TABLE IF NOT EXISTS files (
	path TEXT PRIMARY KEY,
	description TEXT,
	charDate TIMESTAMP)"""
_mainDDLfileindex = """CREATE INDEX IF NOT EXISTS path_index ON files (path)"""
_tagsDDL = """CREATE TABLE IF NOT EXISTS tags (
	tag TEXT,
	path TEXT)"""


def stringEscape(aString):
	"""returns an SQL literal for aString.

	I know there's code to do this in pysqlite -- but where?
	"""
	if isinstance(aString, unicode):
		aString = aString.encode("utf-8")
	return "'%s'"%(aString.replace("'", "''"))

class Annotations(object):
	"""is a facade for Annotations held in an SQLite DB.
	"""
	def __init__(self, storage, filesName="files", tagsName="tags"):
		if isinstance(storage, basestring):
			self.connection = sqlite.connect(storage, 
				detect_types=sqlite.PARSE_DECLTYPES|sqlite.PARSE_COLNAMES)
		else:
			self.connection = storage
		self.filesName, self.tagsName = filesName, tagsName
		self.subCount = 0
		self._setupTables()

	def __contains__(self, key):
		res = self.connection.execute("SELECT path FROM %s WHERE path=?"%
			self.filesName, (key,)).fetchall()
		if res:
			return True
		return False

	def __iter__(self):
		for key in self.keys():
			yield self.get(key)

	def _setupTables(self):
		self.connection.execute(_mainDDL)
		self.connection.execute(_mainDDLfileindex)
		self.connection.execute(_tagsDDL)
	
	def save(self):
		self.connection.commit()

	def delete(self, key, cursor=None):
		if cursor is None:
			cursor = self.connection.cursor()
		for tableName in [self.filesName, self.tagsName]:
			cursor.execute("DELETE FROM %s WHERE path=?"%tableName, (key,))

	def keys(self):
		return [r[0] for r in self.connection.execute(
			"SELECT path FROM %s WHERE charDate IS NOT NULL"
				" ORDER BY charDate"%self.filesName)]

	def add(self, ann):
		path, description, charDate = ann.key, ann.description, ann.charDate
		cursor = self.connection.cursor()
		self.delete(path, cursor=cursor)
		cursor.execute("INSERT INTO %s (path, description, charDate)"
			" VALUES (?,?,?)"%self.filesName, (path, description, charDate))
		cursor.executemany("INSERT INTO %s (path, tag)"
			" VALUES (?,?)"%self.tagsName, [(path, tag) for tag in ann.tags])
		cursor.close()

	def get(self, key):
		res = self.connection.execute("SELECT * FROM %s WHERE path=?"%
			self.filesName, (key,)).fetchall()
		if not res:
			return None
		res = res[0]
		key, description, charDate = res[0], res[1], res[2]
		tags = [r[0] for r in self.connection.execute(
				"SELECT tag FROM %s WHERE path=?"%self.tagsName, (key,))]
		return common.Annotation(key, description, tags, charDate)

	def getTagFreqs(self):
		return dict(self.connection.execute(
			"SELECT tag, count(*) FROM %s GROUP BY tag"%self.tagsName))

	def _makeSelectedTagsView(self, destTableName, tags):
		subTableSelect = " INTERSECT ".join(
			'SELECT path FROM %s WHERE tag=%s'%(self.tagsName,
			stringEscape(tag)) for tag in tags)
		self.connection.execute("CREATE TEMP VIEW %s AS "
			"SELECT tag, path FROM %s "
			"WHERE path IN (%s)"%(destTableName, self.tagsName,
			subTableSelect))

	def getSubCollection(self, tags):
		uPrefix = "v%s_%s"%(str(abs(id(self))), self.subCount)
		self.subCount += 1
		tmpFilesName, tmpTagsName = uPrefix+"files", uPrefix+"tags"
		self._makeSelectedTagsView(tmpTagsName, tags)
		self.connection.execute("""CREATE TEMP VIEW %s AS 
			SELECT 
				files.path AS path, 
				files.description AS description, 
				files.charDate AS charDate
			FROM (SELECT DISTINCT path FROM %s) AS q 
			NATURAL JOIN %s as files"""%(
			tmpFilesName, tmpTagsName, self.filesName))
		return Annotations(self.connection, tmpFilesName, tmpTagsName)

	def iterDatesKeys(self):
		return self.connection.execute("SELECT charDate, path FROM %s"
				" WHERE charDate IS NOT NULL ORDER BY charDate"%self.filesName)


def importOld():
	ann = Annotations("/home/msdemlei/.ianno.sqlite")
	import txannotations
	oldAnns = txannotations.Annotations()
	for k in oldAnns.keys():
		if not k in ann:
			oa = oldAnns.get(k)
			oa.updateDate()
			if oa.charDate is not None:
				ann.add(oa)
	ann.save()


if __name__=="__main__":
	importOld()
