#!/usr/bin/env python

import os
import time
import Tkinter

import Image
import ImageTk
import PIL.ImageOps as ImageOps

from ianno import annotate
from ianno import common
from ianno import config
from ianno import dbannotations
from ianno import tagcloud
from ianno import timeline


class CloudControl(tagcloud.TagCloud):
	def _buildWidgets(self):
		tagcloud.TagCloud._buildWidgets(self)
		buttonFrame = Tkinter.Frame(self)
		self.updateButton = Tkinter.Button(buttonFrame, text="Update")
		self.updateButton.pack(side=Tkinter.LEFT)
		self.resetButton = Tkinter.Button(buttonFrame, text="Reset")
		self.resetButton.pack(side=Tkinter.LEFT)
		buttonFrame.grid(row=3, column=1, columnspan=2, sticky=Tkinter.E)


class CanvasImage(object):
	"""A wrapper for the image displayed.

	It will handle all kinds of corner cases: Images that are deleted, blanked
	images, etc.

	Basically, it manages one id on the parent canvas, which can be either
	an info text or an image.  By default, it's always centered on the canvas'
	center.

	To keep it that way, it monitors the parent's configure requests.  In
	order to avoid scaling oodles of times, it will only honor them with
	some delay (it would be nicer to peek into the event queue, but I've
	not tried to figure out how that works).
	"""
	font = ("Helvetica", 25)
	mouseBoost = 4

	def __init__(self, canvas):
		self.canvas = canvas
		self.managedId = None    # canvas id of what we're displaying
		self.configAfter = None  # after id of reconfiguration call
		self.image = None        # image as loaded, if present
		self.zoom = "fit"        # image scale, valid keys in scalers
		self.drawText("No image")
		self._doBindings()
	
	def _doBindings(self):
		self.canvas.bind("<Configure>", self._handleConfigure)
		self.canvas.bind('<Button-1>', self._startMove)
		self.canvas.bind('<B1-Motion>', self._doMove)
		self.canvas.bind('<ButtonRelease-1>', self._endMove)

	def _handleConfigure(self, ev):
		# see if there has been a configure request lately and cancel it if so
		if self.configAfter is not None:
			self.canvas.after_cancel(self.configAfter)
		self.configAfter = self.canvas.after(100, self._reconfigure)

	def _reconfigure(self):
		# actually reconfigure; this is only called through an after to
		# make sure we're not redrawing to frequently
		self.canvas.coords(self.managedId, *self._getCenter())
		if self.zoom=="fit" and self.image:
			self._setScaledImage()

	def _clear(self):
		if self.managedId is not None:
			self.canvas.delete(self.managedId)

	def _getCenter(self):
		return self.canvas.winfo_width()/2, self.canvas.winfo_height()/2

	def _startMove(self, event):
		self._lastx, self._lasty = event.x, event.y

	def _doMove(self, event):
		self.canvas.move(self.managedId, (event.x-self._lastx)*self.mouseBoost,
			(event.y-self._lasty)*self.mouseBoost)
		self._lastx, self._lasty = event.x, event.y

	def _endMove(self, event):
		self.canvas.move(self.managedId, (event.x-self._lastx)*self.mouseBoost,
			(event.y-self._lasty)*self.mouseBoost)
		del self._lastx
		del self._lasty

	def _updateDisplayedImage(self, image):
		"""clears the display and makes the PIL image a new canvas image.
		"""
		self._clear()
		# We need to keep the reference since Canvas won't
		self._displayedImage = ImageTk.PhotoImage(image)
		x, y = self._getCenter()
		self.managedId = self.canvas.create_image(x, y,
			image=self._displayedImage)

	def _setScaledImage(self):
		self._updateDisplayedImage(
			common.fitImage(self.image, self.canvas))

	def _setNormalImage(self):
		self._updateDisplayedImage(self.image)

	def _setDoubleImage(self):
		iw, ih = self.image.size
		self._updateDisplayedImage(self.image.resize(
			(iw*2, ih*2), Image.ANTIALIAS))

	scalers = {
		"single": _setNormalImage,
		"double": _setDoubleImage,
		"fit": _setScaledImage,
	}

	def setZoom(self, zoom):
		"""sets the current zoom.

		Only keys from self.scalers are allowed, but we don't check this
		(since it's supposed to be an internal error).
		"""
		if self.image:
			self.scalers[zoom](self)
		self.zoom = zoom

	def setImage(self, path):
		try:
			self.image = Image.open(path)
		except IOError, msg:
			self.drawText("Could not open image %s\n(%s)"%(path, str(msg)))
			self.image = None
		if self.image:
			self.scalers[self.zoom](self)

	def drawText(self, text):
		self._clear()
		x, y = self._getCenter()
		self.managedId = self.canvas.create_text(x, y,
			text=text, font=self.font, fill="#FFFFFF")
		

class ImageViewer(Tkinter.Tk):
	"""A Tkinter GUI for viewing annotated images.
	"""
	current = None

	def __init__(self, annotations, initTags=[]):
		Tkinter.Tk.__init__(self)
		self.geometry(config.get("iaview", "viewerGeometry"))
		self.fullAnnotations = annotations
		if initTags:
			self.annotations = self.fullAnnotations.getSubCollection(initTags)
		else:
			self.annotations = self.fullAnnotations
		self._newImageListeners = []
		self._makeWidgets()
		self._doBindings()
		self.wait_visibility(self.canvas)
		self._setFList(self.annotations.keys())
		if config.get("iaview", "fulled"):
			self._toggleFullscreen()

	def _makeWidgets(self):
		self.canvas = Tkinter.Canvas(self, background="black")
		self.image = CanvasImage(self.canvas)
		self.canvas.pack(fill=Tkinter.BOTH, expand=True)
		self._createTagCloud()
		self._createTimeline()

	def _createTagCloud(self):
		self.tagCloudDialogue = Tkinter.Toplevel(self)
		self.tagCloudDialogue.title("iaview Tag Cloud")
		self.tagCloudDialogue.withdraw()
		self.tagCloud = CloudControl(self.tagCloudDialogue, 
			self.annotations.getTagFreqs())
		self.tagCloud.pack(expand=True, fill=Tkinter.BOTH)
		self.tagCloudDialogue.geometry(config.get("iaview", "tagGeometry"))
		self.tagCloud.updateButton.config(command=self._updateCloud)
		self.tagCloud.resetButton.config(command=self._resetCloud)

	def _createTimeline(self):
		self.timelineDialogue = Tkinter.Toplevel(self)
		self.timelineDialogue.title("iaview Time Line")
		self.timelineDialogue.withdraw()
		self.timeline = timeline.Timeline(self.timelineDialogue)
		self.timeline.pack(expand=True, fill=Tkinter.BOTH)
		self.timelineDialogue.geometry(config.get("iaview", "timelineGeometry"))
		self.timeline.addClickListener(self._setCurrentFromKey)
		self.addNewImageListener(self.timeline.notifySelection)

	def _doBindings(self):
		self.bind_all("q", lambda ev: self.quit())
		self.bind("f", lambda ev: self._toggleFullscreen())
		self.bind("1", lambda ev: self.image.setZoom("single"))
		self.bind("2", lambda ev: self.image.setZoom("double"))
		self.bind("s", lambda ev: self.image.setZoom("fit"))
		self.bind_all("a", lambda ev: self._showDescription())
		self.bind_all("y", lambda ev: self._yankKey())
		self.bind_all("t", lambda ev: self._toggleTagCloud())
		self.bind_all("d", lambda ev: self._toggleTimeline())
		self.bind_all("<space>", lambda ev: self._next())
		self.bind_all("<BackSpace>", lambda ev: self._prev())
		self.focus()

	def _setCurrentFromKey(self, key):
		index = self.fList.index(key) # linear search doesn't hurt here,
		                              # since img update is much slower.
		self._setCurrent(index)

	def _setCurrent(self, newIndex):
		key = self.fList[newIndex]
		self.ann = self.annotations.get(key)
		self.current = newIndex
		self.image.setImage(key)
		self._broadcastNewImage(key)

	def _setFList(self, fList):
		self.fList = fList
		self._updateForAnnotations()
		if self.fList:
			self._setCurrent(0)
		else:
			self.image.drawText("No images selected")

	def _next(self):
		if self.current<len(self.fList)-1:
			self._setCurrent(self.current+1)

	def _prev(self):
		if self.current>0:
			self._setCurrent(self.current-1)

	def _toggleTagCloud(self):
		if self.tagCloudDialogue.state()=='withdrawn':
			self.tagCloudDialogue.deiconify()
		else:
			self.tagCloudDialogue.withdraw()

	def _toggleTimeline(self):
		if self.timelineDialogue.state()=='withdrawn':
			self.timelineDialogue.deiconify()
		else:
			self.timelineDialogue.withdraw()

	def _showDescription(self):
		if self.ann:
			id = self.canvas.create_window(self.canvas.canvasx(0), 
				self.canvas.canvasy(0), anchor=Tkinter.NW, 
				window=Tkinter.Label(text=self.ann.description, wraplength=200,
					justify=Tkinter.LEFT))
			self.after(3000, lambda: self.canvas.delete(id))

	def _returnSelection(self, *args):
		return self.selection

	def _yankKey(self):
		try:
			key = self.fList[self.current]
			self.selection = key
			self.clipboard_clear()
			self.clipboard_append(key)
			self.selection_own(selection="PRIMARY")
			self.selection_handle(self._returnSelection, selection="PRIMARY")
		except IndexError:
			pass

	def _setFullscreen(self):
# Some day, use the _NET_WM_STATE_FULLSCREEN property or:
#		self.wm_state("zoomed")
		self.prevGeom = self.geometry()
		os.system("wmctrl -i -r %s -b add,fullscreen"%self.frame())

	def _setWindowed(self):
		del self.prevGeom
		os.system("wmctrl -i -r %s -b remove,fullscreen"%self.frame())

	def _toggleFullscreen(self):
		if getattr(self, "prevGeom", None):
			self._setWindowed()
		else:
			self._setFullscreen()
	
	def _updateCloud(self):
		selTags = self.tagCloud.getSelectedItems()
		self.annotations = self.annotations.getSubCollection(selTags)
		self._setFList(self.annotations.keys())
		self.tagCloud.select(selTags)
	
	def _resetCloud(self):
		self.annotations = self.fullAnnotations
		self._setFList(self.annotations.keys())
	
	def _updateForAnnotations(self):
		self.tagCloud.setFreqs(self.annotations.getTagFreqs())
		self.tagCloud.layout()
		self.timeline.setItems(self.annotations.iterDatesKeys())

	def _broadcastNewImage(self, key):
		for listener in self._newImageListeners:
			listener(key)

	def addNewImageListener(self, listener):
		"""registers listener for notification when a new image is being
		displayed.

		The argument is the image key.  It may be None if not current image
		is available.
		"""
		self._newImageListeners.append(listener)

	def getVolatiles(self):
		if getattr(self, "prevGeom", None):
			yield "viewerGeometry", self.prevGeom
			yield "fulled", "True"
		else:
			yield "viewerGeometry", self.geometry()
			yield "fulled", "False"
		if self.tagCloudDialogue.state()=="normal":
			yield "tagGeometry", self.tagCloudDialogue.geometry()
		if self.timelineDialogue.state()=="normal":
			yield "timelineGeometry", self.timelineDialogue.geometry()


def main():
	import sys
	iv = ImageViewer(dbannotations.Annotations(
			os.path.join(os.environ["HOME"], ".ianno.sqlite")), sys.argv[1:])
	iv.title("iaview")
	iv.mainloop()
	for key, value in iv.getVolatiles():
		config.set("iaview", key, value)
	config.save()

if __name__=="__main__":
	main()
