"""
A simple tag cloud widget for Tkinter
"""


import bisect
import math
import Tkinter


class TagCloud(Tkinter.Frame):
	minHSpace = 5
	rightMargin = 5
	minFontSize = 7
	freqBoost = 3
	activeColor = "#FF0000"

	def __init__(self, master, tagFreqs, **kwargs):
		Tkinter.Frame.__init__(self, master, **kwargs)
		self._buildWidgets()
		self.setFreqs(tagFreqs)
		self.canvas.bind("<Configure>", lambda ev: self.layout())
		self.canvas.bind("<Map>", lambda ev: self.layout())
		self.canvas.tag_bind("inactive", "<1>", self._selectCurrent)
		self.canvas.tag_bind("active", "<1>", self._unselectCurrent)

	def setFreqs(self, tagFreqs):
		self.canvas.delete(Tkinter.ALL)
		self.items, self.tagMap, self.itemProperties = {}, {}, []
		self.tagFreqs = tagFreqs
		self._makeItems()

	def addTag(self, tag, freq=1):
		self.tagFreqs[tag] = freq
		self._makeItem(tag)
		self.layout()
		self.select(tag)

	def getSelectedItems(self):
		return set(self.items[i] for i in self.canvas.find_withtag("active"))

	def select(self, tags):
		if isinstance(tags, basestring):
			tags = [tags]
		for tag in tags:
			self._select(self.tagMap[tag])

	def clear(self):
		for i in self.canvas.find_withtag("active"):
			self._unselect(i)

	def _makeItem(self, tag):
		fontSize = int(self.minFontSize+self.freqBoost*math.log(
			self.tagFreqs[tag]+1))
		curItem = self.canvas.create_text(0, 0, text=tag, 
			anchor=Tkinter.NW, font=('Helvetica', fontSize), tags=("inactive",))
		self.items[curItem] = tag
		self.tagMap[tag] = curItem
		x0, y0, x1, y1 = self.canvas.bbox(curItem)
		tw, th = x1-x0, y1-y0
		bisect.insort(self.itemProperties, (tag, curItem, tw, th))

	def _makeItems(self):
		self.items, self.tagMap, self.itemProperties = {}, {}, []
		for tag in self.tagFreqs:
			self._makeItem(tag)

	def _select(self, item):
		self.canvas.itemconfig(item, tags=("active",),
			fill=self.activeColor)

# Tkinter.CURRENT requires the mouse to enter and exit the item between
# two firings at least with python 2.4.  So, let's work around this.
	def _selectCurrent(self, ev):
		if self.canvas.find_withtag(Tkinter.CURRENT):
			self._select(Tkinter.CURRENT)
		else:
			self._select(self.canvas.find_closest(self.canvas.canvasx(ev.x), 
				self.canvas.canvasy(ev.y)))

	def _unselect(self, item):
		self.canvas.itemconfig(item, tags=("inactive",),
			fill="#000000")

	def _unselectCurrent(self, ev):
		if self.canvas.find_withtag(Tkinter.CURRENT):
			self._unselect(Tkinter.CURRENT)
		else:
			self._unselect(
				self.canvas.find_closest(self.canvas.canvasx(ev.x), 
					self.canvas.canvasy(ev.y)))

	def _makeCanvasFrame(self):
		canvasFrame = Tkinter.Frame(self)
		self.canvas = Tkinter.Canvas(canvasFrame, background="#FFFFFF")
		canScroll = Tkinter.Scrollbar(canvasFrame, command=self.canvas.yview,
			width=10)
		self.canvas.config(yscrollcommand=canScroll.set)
		self.canvas.grid(column=1, row=0, sticky=Tkinter.NW+Tkinter.SE)
		canScroll.grid(column=2, row=0, sticky=Tkinter.N+Tkinter.S)
		canvasFrame.columnconfigure(1, weight=1, minsize=100)
		canvasFrame.rowconfigure(0, weight=1, minsize=150)
		return canvasFrame

	def _buildWidgets(self):
		self._makeCanvasFrame().grid(column=1, row=2, columnspan=2, 
			sticky=Tkinter.NW+Tkinter.SE)
		self.columnconfigure(1, weight=1)
		self.rowconfigure(2, weight=1)

	def layout(self):
		canvasWidth = self.canvas.winfo_width()
		if canvasWidth<10 or not self.itemProperties:
			return

		def collectALine(firstItem, items):
			itemsInLine = [firstItem]
			curWidth, curHeight = firstItem[2], firstItem[3]
			try:
				while True:
					item = items.next()
					_, _, width, height = item
					if curWidth+self.minHSpace+width>canvasWidth:
						return item, itemsInLine, curHeight
					else:
						itemsInLine.append(item)
						curHeight = max(curHeight, height)
						curWidth += self.minHSpace+width
			except StopIteration:
				pass
			return None, itemsInLine, curHeight

		def layoutLine(itemsInLine, cury):
			curx = 0
			if len(itemsInLine)<2:
				curHSpace = 0
			else:
				curHSpace = (canvasWidth-sum([i[2] for i in itemsInLine]))/(
					len(itemsInLine)-1)
			for _, id, width, _ in itemsInLine:
				self.canvas.coords(id, curx, cury)
				curx += width+curHSpace

		cury = 0
		items = iter(self.itemProperties)
		lastItem = items.next()
		while lastItem:
			lastItem, itemsInLine, lineHeight = collectALine(lastItem, items)
			layoutLine(itemsInLine, cury)
			cury += lineHeight
		self.canvas.config(scrollregion=self.canvas.bbox(Tkinter.ALL))


class TagCloudWithAdder(TagCloud):
	def _buildWidgets(self):
		TagCloud._buildWidgets(self)
		tagEntry = Tkinter.Entry(self)
		def addAction():
			newTag = tagEntry.get().strip()
			if newTag:
				self.addTag(newTag)
			tagEntry.delete(0, Tkinter.END)
			return "break"
		addButton = Tkinter.Button(self, text="Add", command=addAction)
		tagEntry.bind("<Return>", lambda ev: addAction())
		tagEntry.grid(column=1, row=1, sticky=Tkinter.NW+Tkinter.SE)
		addButton.grid(column=2, row=1)


def _makeTestSet(n):
	import random
	def makeRandomWord():
		chars = [chr(random.randint(65, 90)+32)]
		while random.random()>0.2:
			chars.append(chr(random.randint(65, 90)+32))
		return "".join(chars)
	for i in range(n):
		yield (makeRandomWord(), random.randint(0, 20)**2+random.randint(0, 100))


if __name__=="__main__":
	root = Tkinter.Tk()
	tc = TagCloud(root, dict(_makeTestSet(100)))
	tc.pack(expand=1, fill=Tkinter.BOTH)
	tc.bind("q", lambda ev: tc.quit())
	tc.focus()
	root.mainloop()
