"""
A time-line widget for Tkinter.

Basically, you register a number of "items" consisting of a datatime.datetime
object and a user object (setItems).  The user object must be hashable.

Then you listen to clicked events (addClickListener).  The function passed gets
called with the (datetime, user object) tuple if the user clicks on an item.

Once you've reacted on such an event, call notifySelection with the
relevant user object.
"""

# XXX TODO: Pan and Zoom.

import bisect
import datetime
import Tkinter

from math import log10 as log


def timedeltaToSecs(timeDelta):
	return timeDelta.days*24*3600+timeDelta.seconds+timeDelta.microseconds/1e6


class TimeLabelIterator(object):
	"""TimeLabelIterators know how to yield "sensible" labels for time-type
	axes.

	You have to give _intervals the iterator feels reponsible for and an
	_intUnit factor giving their unit (in seconds); _intervals is supposed
	to be sorted in descending order, and a _labelFormat as well as 
	_titleFromat (strftime formats).

	Further, you have to provide

	* a method _iterDts(startDt, interval) yielding datetimes of 
	  approximately the desired intervals.  interval is one of the choices
	  from your _intervals, startDt is whatever you returned from
	  _getStartDt (unless you override iterLabels).
	* a method _getStartDt(startDt, interval) returning an appopriate
	  first label and a boolean saying whether this label is within
	  the interval asked for (unless you override iterLabels).
	
	TimeLabelIterators have an attribute oomCenter giving the mean of the
	logs of the number of seconds in the shortest and longest intervals.
	This is intended in helping to figure out what TimeLabelIterator to
	use for some given interval.
	"""
	def __init__(self, labelDistMin=40):
		self.labelDistMin = labelDistMin
		self._realIntervals = [self._intUnit*i for i in self._intervals]
		self.oomCenter = log((self._realIntervals[0]+self._realIntervals[-1])/2)

	def _chooseSecondsInterval(self, totalSeconds, pixelWidth):
		"""returns an interval out of _realIntervals that the function deems
		optimal for covering totalSeconds on pixelWidth pixels.

		"Optimal" is the largest number of labels without having them closer
		together than labelDistMin.  A special rule is that we'd rather have
		smashed labels than only one, so we special-case that.
		"""
		for index, dist in enumerate(self._realIntervals):
			numLabels = totalSeconds/float(dist)
			labelDist = pixelWidth/numLabels
			if labelDist>self.labelDistMin:
				break
		if numLabels<1.8 and index:  # Don't have only one label if you can help it
			return self._realIntervals[index-1]
		return dist

	def iterLabels(self, startDt, totalSeconds, pixelWidth):
		interval = self._chooseSecondsInterval(totalSeconds, pixelWidth)
		endDt = startDt+datetime.timedelta(seconds=totalSeconds)
		startDt, dropFirst = self._getStartDt(startDt, interval)
		iter = self._iterDts(startDt, interval)
		if dropFirst:
			iter.next()
		for dt in iter:
			if dt>endDt:
				break
			yield dt, dt.strftime(self._labelFormat)

	def getTitle(self, startDt):
		return startDt.strftime(self._titleFormat)


class SecondsLabelIterator(TimeLabelIterator):
	"""is an iterator providing labels for second-like intervals.

	It doubles as a base class for iterators being able to iterate on
	plain seconds intervals (like Minute or Hour).
	"""
	_intervals = [1, 2, 5, 10, 15, 30, 60]
	_intUnit = 1
	_titleFormat = "%H:%M"
	_labelFormat = ":%M:%S"

	def _getStartDt(self, startDt, interval):
		secs = startDt.second
		return datetime.datetime(*(startDt.timetuple()[:5]+(secs-secs%interval,))
			), secs%interval

	def _iterDts(self, startDt, interval):
		increment = datetime.timedelta(seconds=interval)
		while True:
			yield startDt
			startDt = startDt+increment


class MinutesLabelIterator(SecondsLabelIterator):
	_intUnit = 60
	_titleFormat = "%Y-%m-%d"
	_labelFormat = "%H:%M"

	def _getStartDt(self, startDt, interval):
		mins = startDt.minute
		return datetime.datetime(*(startDt.timetuple()[:4]+(
			mins-mins%(interval/60),))), (mins%(interval/60) or 
				startDt.second!=0)


class HoursLabelIterator(SecondsLabelIterator):
	_intervals = [1, 2, 4, 8, 12, 24]
	_intUnit = 3600
	_titleFormat = "%Y-%m-%d"
	_labelFormat = "%Hh"

	def _getStartDt(self, startDt, interval):
		hours = startDt.hour
		return datetime.datetime(*(startDt.timetuple()[:3]+(
			hours-hours%(interval/3600),))), (hours%(interval/3600) or 
				startDt.minute!=0)


class DaysLabelIterator(SecondsLabelIterator):
	_intervals = [1, 2, 5, 10, 20, 30]
	_intUnit = 3600*24
	_titleFormat = "%Y-%m-"
	_labelFormat = "%d"
	
	def _getStartDt(self, startDt, interval):
		days = startDt.day
		interval = interval/self._intUnit
		return datetime.datetime(*(startDt.timetuple()[:2]+(max(1, 
			days-days%interval),))), ((days%interval) or startDt.hour!=0)
	
	def _iterDts(self, startDt, interval):
		increment = datetime.timedelta(seconds=interval)
		interval = interval/self._intUnit
		while True:
			yield startDt
			startDt = startDt+increment
			if startDt.day%interval:
				startDt = datetime.datetime(startDt.year, startDt.month,
					interval)


class MonthsLabelIterator(TimeLabelIterator):
	_intervals = [1, 2, 3, 6, 12]
	_intUnit = 3600*24*30 # about a month
	_titleFormat = "%Y"
	_labelFormat = "%m"

	def _getStartDt(self, startDt, interval):
		return datetime.datetime(startDt.year, startDt.month, 1), startDt.day!=1
	
	def _iterDts(self, startDt, interval):
		interval = interval/self._intUnit
		while True:
			yield startDt
			nextMonth = startDt.month+interval
			if nextMonth>12:
				startDt = datetime.datetime(startDt.year+1, nextMonth-12, 1)
			else:
				startDt = datetime.datetime(startDt.year, nextMonth, 1)


class YearsLabelIterator(TimeLabelIterator):
	_intervals = [1, 2, 5, 10, 20, 25, 50, 100]
	_intUnit = 3600*24*365 # about a year
	_titleFormat = ""
	_labelFormat = "%Y"

# We disregard the fact that there's no year 0 here.  AD only, please...
	def _getStartDt(self, startDt, interval):
		interval = interval/self._intUnit
		return datetime.datetime(startDt.year-startDt.year%interval, 1, 1
			), startDt.year%interval or startDt.month!=1

	def _iterDts(self, startDt, interval):
		interval = interval/self._intUnit
		while True:
			yield startDt
			startDt = datetime.datetime(startDt.year+interval, 1, 1)


class LabelIteratorChooser(object):
	"""is a dispatcher for the XLabelIterators.
	
	This is intended to be constructed once and then be reused every time
	you need to choose a LabelIterator.
	"""
	def __init__(self, labelDistMin=30,
		iteratorClasses=[SecondsLabelIterator, 
			MinutesLabelIterator, HoursLabelIterator, DaysLabelIterator, 
			MonthsLabelIterator, YearsLabelIterator]):
		self.iterators = [(it.oomCenter, it) for it in 
			[cls(labelDistMin) for cls in iteratorClasses]]
	
	def getIterator(self, interval):
		logInt = log(interval)
		oomOffsets = [(abs(center-logInt), it) 
			for center, it in self.iterators]
		return min(oomOffsets)[1]


class Timeline(Tkinter.Frame):
	selColor = "red"
	unselColor = "green"
	xMargin = 2
	yMargin = 2

	def __init__(self, parent, **kwargs):
		if "selColor" in kwargs:
			self.selColor = kwargs.pop("selColor")
		if "unselColor" in kwargs:
			self.unselColor = kwargs.pop("unselColor")
		self.selKey = None
		Tkinter.Frame.__init__(self, parent, **kwargs)
		self.labelChooser = LabelIteratorChooser()
		self._listeners = []
		self.userItems = []
		self._makeWidgets()
		self._clearCanvas()
		self.canvas.bind("<1>", self._handleClick)
	
	def _makeWidgets(self):
		self.canvas = Tkinter.Canvas(self)
		self.canvas.pack(expand=True, fill=Tkinter.BOTH)
		self.canvas.bind("<Configure>", self._layout)

	def _clearCanvas(self):
		self.canvas.delete(Tkinter.ALL)
		self.xToData, self.dataToId = {}, {}
		self.xPositions = []
	
	def _makeItems(self, height):
		for timestamp, userData in self.userItems:
			xLocation = self.dateToX(timestamp)
			if xLocation not in self.xToData:
				self.xToData[xLocation] = userData
				self.xPositions.append(xLocation)
				itemId = self.canvas.create_line(xLocation, 0, xLocation, height,
					fill=self.unselColor)
			self.dataToId[userData] = itemId

	def _makeLabels(self, startDt, totalSeconds, pixelWidth, height):
		labelMaker = self.labelChooser.getIterator(totalSeconds)
		for dt, label in labelMaker.iterLabels(startDt, totalSeconds, pixelWidth):
			self.canvas.create_text(self.dateToX(dt), height-self.yMargin, 
				text=label, anchor=Tkinter.S)
		self.canvas.create_text(self.xMargin, self.yMargin, 
			text=labelMaker.getTitle(startDt), anchor=Tkinter.N)

	def _layout(self, ignored=None):
		self._clearCanvas()
		width, height = self.winfo_width(), self.winfo_height()
		if not width:  # we're not mapped yet
			return
		if not self.userItems: # no items (yet)
			return
		width = max(self.xMargin*2, width-self.xMargin*2)
		startDt = self.userItems[0][0]
		totalSeconds = max(60, timedeltaToSecs(self.userItems[-1][0]-startDt))

		def getPos(timestamp):
			return self.xMargin+int(
				timedeltaToSecs(timestamp-startDt)/totalSeconds*width)
		self.dateToX = getPos

		self._makeItems(height)
		self._makeLabels(startDt, totalSeconds, width, height)

		if self.selKey is not None:
			self.notifySelection(self.selKey)

	def _broadcastSelection(self, selection):
		for listener in self._listeners:
			listener(selection)

	def _handleClick(self, ev):
		if not self.xPositions: # no points to select
			return
		ind = bisect.bisect_left(self.xPositions, ev.x)
		# ind now points to the smallest x right of the click point.  So,
		# we check ind and the previous one for matches.
		if ind<len(self.xPositions):
			pos0 = pos1 = self.xPositions[ind]
		else:
			pos0 = pos1 = self.xPositions[ind-1]
		if ind:
			pos1 = self.xPositions[ind-1]
		dist, x = min([(abs(ev.x-pos0), pos0), (abs(ev.x-pos1), pos1)])
		if dist<3:
			self._broadcastSelection(self.xToData[x])

	def setItems(self, items):
		self.userItems = list(items)
		self.userItems.sort()
		self.selKey = None
		self._layout()

	def setItemsHavingTimestamps(self, items):
		self.userItems = [(o.timestamp, o) for o in items]
		self.userItems.sort()
		self._layout()

	def notifySelection(self, userData):
		self.canvas.itemconfigure("SELECTED", fill=self.unselColor, tags=())
		if userData not in self.dataToId:
			return
		self.canvas.itemconfig(self.dataToId[userData], tags=("SELECTED",),
			fill=self.selColor)
		self.selKey = userData

	def addClickListener(self, listener):
		self._listeners.append(listener)


def _testWidget():
	import datetime
	import sys
	class UserItem(object):
		def __init__(self, timestamp):
			self.timestamp = timestamp
	root = Tkinter.Tk()
	root.bind("q", lambda ev: root.quit())
	root.focus()
	tl = Timeline(root)
	tl.pack(expand=True, fill=Tkinter.BOTH)
	tl.setItemsHavingTimestamps([
		UserItem(datetime.datetime(2000, 5, 3, 23, 10)),
		UserItem(datetime.datetime(2002, 8, 3, 12, 10)),
		UserItem(datetime.datetime(2002, 8, 3, 12, 12)),
		UserItem(datetime.datetime(2008, 8, 3, 12, 12)),])
	tl.addClickListener(lambda data: tl.notifySelection(data))
	root.mainloop()


def getTestSuite(methodPrefix=None):
	import unittest

	class ScalerTest(unittest.TestCase):
		testedClass = None
		
		def getIterator(self, totalSeconds):
			return self.testedClass()

		def assertLabels(self, startTuple, totalSeconds, pixelWidth, expectation):
			res = [label for dt, label in self.getIterator(totalSeconds).iterLabels(
					datetime.datetime(*startTuple), totalSeconds, pixelWidth)]
			self.assertEqual(res, expectation)

	class SecondsIteratorTest(ScalerTest):
		testedClass = SecondsLabelIterator

		def testTwoSeconds(self):
			self.assertLabels((2000, 12, 20, 10, 10, 12), 10, 400,
				[':10:12', ':10:14', ':10:16', ':10:18', ':10:20', ':10:22'])

		def testOneSecond(self):
			self.assertLabels((2000, 12, 20, 10, 10, 12), 10, 1000,
				[':10:12', ':10:13', ':10:14', ':10:15', ':10:16', ':10:17', 
					':10:18', ':10:19', ':10:20', ':10:21', ':10:22'])
		
		def testFiveSeconds(self):
			self.assertLabels((2000, 12, 20, 10, 10, 12), 15, 150,
				[':10:15', ':10:20', ':10:25'])
			self.assertLabels((2000, 12, 20, 10, 10, 12), 20, 150,
				[':10:20', ':10:30'])

		def testMinLabelDist(self):
			self.assertLabels((2000, 12, 20, 10, 10, 12), 20, 60,
				[':10:20', ':10:30'])
			self.assertLabels((2000, 12, 20, 10, 10, 12), 15, 50,
				[':10:20', ])

	class MinutesIteratorTest(ScalerTest):
		testedClass = MinutesLabelIterator

		def testShort(self):
			self.assertLabels((2000, 12, 20, 10, 10, 12), 1200, 800,
				['10:12', '10:14', '10:16', '10:18', '10:20', '10:22', 
					'10:24', '10:26', '10:28', '10:30'])

		def testShortNarrow(self):
			self.assertLabels((2000, 12, 20, 10, 10, 12), 1200, 100,
				['10:20', '10:30'])

		def testLongWide(self):
			self.assertLabels((2000, 12, 20, 10, 10, 12), 8000, 500,
				['10:15', '10:30', '10:45', '11:00', '11:15', '11:30', '11:45', 
					'12:00', '12:15'])

		def testMedNarrow(self):
			self.assertLabels((2000, 12, 20, 10, 10, 00), 8000, 100,
				['11:00', '12:00'])
			self.assertLabels((2000, 12, 20, 10, 00, 00), 8000, 100,
				['10:00', '11:00', '12:00'])
			self.assertLabels((2000, 12, 20, 10, 00, 00), 5000, 100,
				['10:00', '10:30', '11:00'])

	class HoursIteratorTest(ScalerTest):
		testedClass = HoursLabelIterator

		def testShortNarrow(self):
			self.assertLabels((2000, 12, 20, 10, 10, 00), 3600*3, 100,
				['11h', '12h', '13h'])
			self.assertLabels((2000, 12, 20, 10, 00, 00), 3600*3, 100,
				['10h', '11h', '12h', '13h'])
			self.assertLabels((2000, 12, 20, 10, 00, 00), 3600*6, 100,
				['10h', '12h', '14h', '16h'])

		def testLongNarrow(self):
			self.assertLabels((2000, 12, 20, 10, 10, 00), 3600*30, 100,
				['12h', '00h', '12h'])
			self.assertLabels((2000, 12, 20, 10, 00, 00), 3600*30, 100,
				['12h', '00h', '12h'])
			self.assertLabels((2000, 12, 20, 10, 00, 00), 3600*10, 100,
				['12h', '16h', '20h'])

		def testShortWide(self):
			self.assertLabels((2000, 12, 20, 10, 10, 00), 3600*3, 600,
				['11h', '12h', '13h'])
			self.assertLabels((2000, 12, 20, 10, 00, 00), 3600*3, 600,
				['10h', '11h', '12h', '13h'])
			self.assertLabels((2000, 12, 20, 10, 00, 00), 3600*6, 600,
				['10h', '11h', '12h', '13h', '14h', '15h', '16h'])

		def testLongWide(self):
			self.assertLabels((2000, 12, 20, 10, 10, 00), 3600*30, 600,
				['12h', '16h', '20h', '00h', '04h', '08h', '12h', '16h'])
			self.assertLabels((2000, 12, 20, 10, 00, 00), 3600*30, 600,
				['12h', '16h', '20h', '00h', '04h', '08h', '12h', '16h'])
			self.assertLabels((2000, 12, 20, 8, 00, 00), 3600*32, 600,
				['08h', '12h', '16h', '20h', '00h', '04h', '08h', '12h', '16h'])
			self.assertLabels((2000, 12, 20, 10, 00, 00), 3600*10, 600,
				['10h', '11h', '12h', '13h', '14h', '15h', '16h', '17h', '18h', 
					'19h', '20h'])


	class DaysIteratorTest(ScalerTest):
		testedClass = DaysLabelIterator

		def testLongWide(self):
			self.assertLabels((2000, 12, 20, 10, 10, 00), 3600*24*30, 600,
				['25', '30', '05', '10', '15'])
			self.assertLabels((2000, 12, 20, 10, 10, 00), 3600*24*30, 1200,
				['22', '24', '26', '28', '30', '02', '04', '06', '08', '10', 
					'12', '14', '16', '18'])

		def testStart(self):
			self.assertLabels((2000, 12, 20, 10, 10, 00), 3600*24*5, 600,
				['21', '22', '23', '24', '25'])
			self.assertLabels((2000, 12, 20, 00, 00, 00), 3600*24*5, 600,
				['20', '21', '22', '23', '24', '25'])
			self.assertLabels((2000, 12, 20, 00, 00, 00), 3600*24*5, 200,
				['20', '22', '24'])
			self.assertLabels((2000, 12, 21, 00, 00, 00), 3600*24*5, 200,
				['22', '24', '26'])
			self.assertLabels((2000, 12, 20, 10, 00, 00), 3600*24*5, 200,
				['22', '24'])


	class MonthsIteratorTest(ScalerTest):
		testedClass = MonthsLabelIterator

		def testShortWide(self):
			self.assertLabels((2000, 11, 20, 10, 10, 00), 3600*24*30*5, 600,
				['12', '01', '02', '03', '04'])
			self.assertLabels((2000, 11, 1, 10, 10, 00), 3600*24*30*5, 600,
				['11', '12', '01', '02', '03'])
			self.assertLabels((2001, 1, 1, 10, 10, 00), 3600*24*30*5, 200,
				['01', '03', '05'])

		def testShortNarrow(self):
			self.assertLabels((2001, 2, 1, 10, 10, 00), 3600*24*30*5, 200,
				['02', '04', '06'])


	class YearsIteratorTest(ScalerTest):
		testedClass = YearsLabelIterator

		def testShortWide(self):
			self.assertLabels((2000, 11, 20, 10, 10, 00), 3600*24*365*5, 600,
				['2001', '2002', '2003', '2004', '2005'])

		def testShortNarrow(self):
			self.assertLabels((2000, 11, 20, 10, 10, 00), 3600*24*365*5, 100,
				['2002', '2004'])

		def testLongWide(self):
			self.assertLabels((2000, 11, 20, 10, 10, 00), 3600*24*365*50, 600,
				['2005', '2010', '2015', '2020', '2025', '2030', '2035', '2040', 
					'2045', '2050'])

		def testLongNarrow(self):
			self.assertLabels((2000, 11, 20, 10, 10, 00), 3600*24*365*50, 200,
				['2020', '2040'])

		def testStart(self):
			self.assertLabels((2000, 11, 20, 10, 10, 00), 3600*24*365*10, 400,
				['2002', '2004', '2006', '2008', '2010'])
			self.assertLabels((2000, 1, 20, 10, 10, 00), 3600*24*365*10, 400,
				['2000', '2002', '2004', '2006', '2008', '2010'])
			self.assertLabels((2001, 1, 20, 10, 10, 00), 3600*24*365*100, 400,
				['2020', '2040', '2060', '2080', '2100'])
			self.assertLabels((2000, 1, 20, 10, 10, 00), 3600*24*365*101, 400,
				['2000', '2020', '2040', '2060', '2080', '2100'])


	class AutoScalingTest(ScalerTest):
		testLabels = LabelIteratorChooser(60)

		def getIterator(self, totalSeconds):
			return self.testLabels.getIterator(totalSeconds)

		def testCountingDownWide(self):
			timeSpan = 3600.*24*365*101
			for expected in [
					 ['1920', '1940', '1960', '1980', '2000'],
					 ['1920', '1930', '1940', '1950', '1960', '1970', '1980'],
					 ['1915', '1920', '1925', '1930', '1935', '1940', '1945', '1950', 
					 	'1955'],
					 ['1915', '1920', '1925', '1930', '1935', '1940'],
					 ['1915', '1920', '1925', '1930'],
					 ['1912', '1914', '1916', '1918', '1920', '1922', '1924', '1926'],
					 ['1912', '1914', '1916', '1918', '1920', '1922'],
					 ['1911', '1912', '1913', '1914', '1915', '1916', '1917', '1918'],
					 ['1911', '1912', '1913', '1914', '1915', '1916'],
					 ['09', '03', '09', '03', '09', '03', '09', '03'],
					 ['09', '03', '09', '03', '09'],
					 ['06', '09', '12', '03', '06', '09', '12', '03'],
					 ['05', '07', '09', '11', '01', '03', '05', '07'],
					 ['05', '07', '09', '11', '01', '03'],
					 ['04', '05', '06', '07', '08', '09', '10', '11'],
					 ['04', '05', '06', '07', '08', '09'],
					 ['04', '05', '06', '07'],
					 ['04', '05', '06'],
					 ['04', '05'],
					 ['25', '30', '05', '10', '15', '20', '25', '30'],
					 ['25', '30', '05', '10', '15', '20'],
					 ['25', '30', '05', '10'],
					 ['24', '26', '28', '30', '02', '04', '06'],
					 ['24', '26', '28', '30', '02'],
					 ['24', '25', '26', '27', '28', '29', '30'],
					 ['24', '25', '26', '27', '28'],
					 ['24', '25', '26'],
					 ['16h', '00h', '08h', '16h', '00h', '08h', '16h'],
					 ['16h', '00h', '08h', '16h', '00h'],
					 ['16h', '20h', '00h', '04h', '08h', '12h', '16h'],
					 ['14h', '16h', '18h', '20h', '22h', '00h', '02h', '04h', '06h', 
					 	'08h'],
					 ['14h', '16h', '18h', '20h', '22h', '00h', '02h'],
					 ['13h', '14h', '15h', '16h', '17h', '18h', '19h', '20h', '21h', 
					 	'22h'],
					 ['13h', '14h', '15h', '16h', '17h', '18h', '19h'],
					 ['13h', '14h', '15h', '16h', '17h'],
					 ['13h', '14h', '15h'],
					 ['12:30', '12:45', '13:00', '13:15', '13:30', '13:45', '14:00', 
					 	'14:15', '14:30'],
					 ['12:30', '12:40', '12:50', '13:00', '13:10', '13:20', '13:30', 
					 	'13:40', '13:50', '14:00'],
					 ['12:30', '12:40', '12:50', '13:00', '13:10', '13:20', '13:30'],
					 ['12:25', '12:30', '12:35', '12:40', '12:45', '12:50', '12:55', 
					 	'13:00', '13:05', '13:10'],
					 ['12:25', '12:30', '12:35', '12:40', '12:45', '12:50', '12:55'],
					 ['12:25', '12:30', '12:35', '12:40', '12:45'],
					 ['12:24', '12:26', '12:28', '12:30', '12:32', '12:34', '12:36', 
					 	'12:38', '12:40'],
					 ['12:24', '12:26', '12:28', '12:30', '12:32', '12:34'],
					 ['12:24', '12:25', '12:26', '12:27', '12:28', '12:29', '12:30', 
					 	'12:31'],
					 ['12:24', '12:25', '12:26', '12:27', '12:28', '12:29'],
					 ['12:24', '12:25', '12:26', '12:27'],
					 [':24:00', ':24:30', ':25:00', ':25:30', ':26:00'],
					 [':23:45', ':24:00', ':24:15', ':24:30', ':24:45', ':25:00', 
					 	':25:15', ':25:30'],
					 [':23:50', ':24:00', ':24:10', ':24:20', ':24:30', ':24:40', 
					 	':24:50', ':25:00'],
					 [':23:50', ':24:00', ':24:10', ':24:20', ':24:30'],
					 [':23:45', ':23:50', ':23:55', ':24:00', ':24:05', ':24:10', 
					 	':24:15', ':24:20'],
					 [':23:45', ':23:50', ':23:55', ':24:00', ':24:05', ':24:10'],
					 [':23:42', ':23:44', ':23:46', ':23:48', ':23:50', ':23:52', 
					 	':23:54', ':23:56', ':23:58', ':24:00'],
					 [':23:42', ':23:44', ':23:46', ':23:48', ':23:50', ':23:52', 
					 	':23:54'],
					 [':23:42', ':23:43', ':23:44', ':23:45', ':23:46', ':23:47', 
					 	':23:48', ':23:49', ':23:50', ':23:51'],
					 [':23:42', ':23:43', ':23:44', ':23:45', ':23:46', ':23:47', 
					 	':23:48'],
					 [':23:42', ':23:43', ':23:44', ':23:45', ':23:46'],
					]:
				self.assertLabels((1910, 3, 23, 12, 23, 42), timeSpan, 600,
					expected)
				timeSpan = timeSpan*0.7

	l = locals()
	if methodPrefix:
		suite = unittest.makeSuite(AutoScalingTest, methodPrefix)
	else:
		tests = [l[name] for name in l 
			if isinstance(l[name], type) and issubclass(l[name], unittest.TestCase)]
		loader = unittest.TestLoader()
		suite = unittest.TestSuite([loader.loadTestsFromTestCase(t)
			for t in tests])
	return suite


def _test():
	import sys, unittest
	if len(sys.argv)>1:
		suite = getTestSuite(sys.argv[1])
	else:
		suite = getTestSuite()
	unittest.TextTestRunner().run(suite)


if __name__=="__main__":
	_testWidget()
	#_test()
